import Control.Monad (liftM, liftM2, (<=<))

{-
  Mini Course: Functional Programming
  Jeremie Gillet

  Day 2 - Pattern Matching
-}

isOne :: Int -> Bool
isOne 1 = True
isOne _ = False

swap :: (a, b) -> (b, a)
swap (a, b) = (b, a)

first :: (a, b, c) -> a
first (a, _, _) = a

second :: (a, b, c) -> b
second (_, b, _) = b

third :: (a, b, c) -> c
third (_, _, c) = c

empty :: [a] -> Bool
empty [] = True
empty _ = False

startsWithX :: String -> Bool
startsWithX ('X' : _) = True
startsWithX _ = False

startsWith :: Char -> String -> Bool
startsWith _ "" = False
startsWith char (x : _) = char == x

true :: a -> Bool
true _ = True

firstFirst :: [(a, b)] -> a
firstFirst [] = error "Empty List"
firstFirst ((x, _) : _) = x

{-
  Day 2 - Recursion
-}

length' :: [a] -> Int
length' [] = 0
length' (_ : xs) = 1 + length' xs

head' :: [a] -> a
head' [] = error "Empty list"
head' (x : _) = x

last' :: [a] -> a
last' [] = error "Empty list"
last' [x] = x
last' (_ : xs) = last' xs

elem' :: (Eq a) => a -> [a] -> Bool
elem' _ [] = False
elem' a (x : xs) = (a == x) || elem' a xs

find :: (a -> Bool) -> [a] -> a
find _ [] = error "Element not found"
find f (x : xs) = if f x then x else find f xs

sum' :: Num a => [a] -> a
sum' [] = 0
sum' (x : xs) = x + sum' xs

maximum' :: Ord a => [a] -> a
maximum' [] = error "Empty list"
maximum' [x] = x
maximum' (x : xs) = max x (maximum' xs)

replicate' :: Int -> a -> [a]
replicate' n x =
  if n <= 0
    then []
    else x : replicate' (n -1) x

take' :: Int -> [a] -> [a]
take' _ [] = []
take' n (x : xs) =
  if n <= 0
    then []
    else x : take' (n -1) xs

all' :: (a -> Bool) -> [a] -> Bool
all' _ [] = True
all' f (x : xs) = f x && all' f xs

any' :: (a -> Bool) -> [a] -> Bool
any' _ [] = False
any' f (x : xs) = f x || any' f xs

map' :: (a -> b) -> [a] -> [b]
map' _ [] = []
map' f (x : xs) = f x : map' f xs

filter' :: (a -> Bool) -> [a] -> [a]
filter' _ [] = []
filter' f (x : xs) =
  let rest = filter' f xs
   in if f x then x : rest else rest

zip' :: [a] -> [b] -> [(a, b)]
zip' [] _ = []
zip' _ [] = []
zip' (a : as) (b : bs) = (a, b) : zip' as bs

reverse' :: [a] -> [a]
reverse' [] = []
-- reverse' (x: xs) = reverse' xs ++ [x]
reverse' xs = go [] xs
  where
    go reversed [] = reversed
    go reversed (x : xs) = go (x : reversed) xs

{-
  Day 2 - Folds
-}

length'' :: [a] -> Int
length'' = foldr (\_ n -> n + 1) 0

head'' :: [a] -> a
head'' = foldl1 const

last'' :: [a] -> a
last'' = foldr1 (flip const)

elem'' :: (Eq a) => a -> [a] -> Bool
elem'' a = foldr (\x acc -> x == a || acc) False

sum'' :: Num a => [a] -> a
sum'' = foldr1 (+)

maximum'' :: Ord a => [a] -> a
maximum'' = foldr1 max

all'' :: (a -> Bool) -> [a] -> Bool
all'' f = foldr (&&) True . map f

any'' :: (a -> Bool) -> [a] -> Bool
any'' f = foldr (||) False . map f

map'' :: (a -> b) -> [a] -> [b]
map'' f = foldr (\x list -> f x : list) []

filter'' :: (a -> Bool) -> [a] -> [a]
filter'' f = foldr (\x list -> if f x then x : list else list) []

reverse'' :: [a] -> [a]
reverse'' = foldl (flip (:)) []

{-
    Day 2 - Maybe
-}

safeTail :: [a] -> Maybe a
safeTail [] = Nothing
safeTail [x] = Just x
safeTail (_ : rest) = safeTail rest

-- Using safeTail, show the tail is it exists, otherwise show "There is no tail"
isThereATail :: [Int] -> String
isThereATail x = case safeTail x of
  Just n -> show n
  Nothing -> "There is no tail"

-- Output a list with three times the tail of there is one. Use pattern matching
threeTimeTheTail :: [a] -> Maybe [a]
threeTimeTheTail x = case safeTail x of
  Just a -> Just [a, a, a]
  Nothing -> Nothing

--Same but use liftM
threeTimeTheTail' :: [a] -> Maybe [a]
threeTimeTheTail' = liftM (\a -> [a, a, a]) . safeTail
