import Control.Monad

-- Implement Binary Tree data type and some functions on it
data Tree a = Nil | Node (Tree a) a (Tree a) deriving (Show)


root :: Tree a -> Maybe a
root Nil = Nothing
root (Node _ a _) = Just a


leftChild :: Tree a -> Maybe (Tree a)
leftChild Nil = Nothing
leftChild (Node lc _ _) = Just lc


rightChild :: Tree a -> Maybe (Tree a)
rightChild Nil = Nothing
rightChild (Node _ _ rc) = Just rc


-- Compute the number of nodes in a tree
size :: Tree a -> Int
size Nil = 0
size (Node lc _ rc) = 1 + size lc + size rc



-- All elements of each left subtree should be smaller than a root value and
-- All elements of each right subtree should be bigger than a root value
addElem :: Ord a => Tree a -> a -> Tree a
addElem Nil x = Node Nil x Nil
addElem (Node lc a rc) x = if x > a
                           then Node lc a (addElem rc x)
                           else Node (addElem lc x) a rc


-- Turn a list into a tree
packList :: Ord a => [a] -> Tree a
packList [] = Nil
packList xs = foldl addElem Nil xs



-- Identify either element is in the tree or not
findElem :: (Ord a, Eq a) => Tree a -> a -> Bool
findElem Nil _ = False
findElem (Node lc a rc) x = if x == a
                            then True
                            else if x > a
                                 then findElem rc x
                                 else findElem lc x


-- map, but for trees
tmap :: (a -> b) -> Tree a -> Tree b
tmap _ Nil = Nil
tmap f (Node lc a rc) = Node (tmap f lc) (f a) (tmap f rc)

{-

    Monad Exercises

-}

-- Define some monadic functions


-- Base function
replicate' :: Int -> a -> [a]
replicate' 0 _ = []
replicate' n a = a : replicate' (n-1) a

-- Monadic function
replicateM' :: Monad m => Int -> m a -> m [a]
replicateM' 0 _ = return []
replicateM' n ma = liftM2 (:) ma (replicateM' (n-1) ma)


-- Base function
zipWith' :: (a -> b -> c) -> [a] -> [b] -> [c]
zipWith' _ [] _ = []
zipWith' _ _ [] = []
zipWith' f (x:xs) (y:ys) = (f x y) : (zipWith' f xs ys)

-- Monadic function
zipWithM' :: Monad m => (a -> b -> m c) -> [a] -> [b] -> m [c]
zipWithM' _ [] _ = return []
zipWithM' _ _ [] = return []
zipWithM' f (x:xs) (y:ys) = liftM2 (:) (f x y) (zipWithM' f xs ys)


-- Base function
filter' :: (a -> Bool) -> [a] -> [a]
filter' _ [] = []
filter' f (x:xs) = if f x
                   then x : rest_filtered
                   else rest_filtered
                   where
                    rest_filtered = filter' f xs


-- Monadic function
filterM' :: Monad m => (a -> m Bool) -> [a] -> m [a]
filterM' _ [] = return []
filterM' f (x:xs) = f x >>= ternary (liftM (x:) rest_filtered) (rest_filtered)
                    where
                        ternary a b cond = if cond then a else b
                        rest_filtered = filterM' f xs




