{-
    Mini Course: Functional Programming
    Jeremie Gillet

    Day 1: Syntax examples - Planning a Birthday Party
-}

-- Comments
-- Happy birthday
{-
  We wish you all the best.
  Love,
  Grandpa & Grandma
-}

-- Functions
getOlder age = age + 1

birthYear age year = year - age

-- Strings
jeremie = ['B', 'i', 'r', 't', 'h', 'd', 'a', 'y', ' ', 'b', 'o', 'y']

-- Lists
invited = ["Justin", "Julie", "James", "Jimmy"]

presents = ["fire truck", "water gun"] ++ ["coloring book", "BB gun"]

wrapPresent present = [present]

collectPresents a b = [a, b]

addPresent newPresent presents = newPresent : presents

-- The dot dot notation
allAges age    = [0 .. age]

allEvenAges age = [0, 2 .. age]

-- Type signatures

greeting :: String -> String
greeting name = "Thank you for coming to my party, " ++ name

birthdayWish :: String -> Int -> String
birthdayWish name age = "Happy " ++ show age ++ "th Birthday, " ++ name

numberOfParticipants :: [a] -> Int
numberOfParticipants participants = length participants

-- if then else
isInvited :: Int -> Bool
isInvited numberOfPresents = if numberOfPresents > 1 then True else False

dear :: String -> Bool -> String
dear name male =
  if male
    then "Dear Mr. " ++ name
    else "Dear Ms. " ++ name

blowCandle :: Int -> Int
blowCandle candles =
  if candles > 0
    then candles - 1
    else 0

-- Tuples
collectInformation :: String -> Int -> (String, Int, Bool)
collectInformation name presents = (name, presents, isInvited presents)

-- Pattern matching

getName :: (String, Int, Bool) -> String
getName (name, presents, invited) = name

unwrapPresent :: [a] -> a
unwrapPresent [present] = present

unwrapFirstPresent :: [a] -> a
unwrapFirstPresent [present] = present
unwrapFirstPresent (present : presents) = present

favoritePresent :: String -> Bool
favoritePresent "fire truck" = True
favoritePresent _ = False

pairParticipants :: [String] -> [(String, String)]
pairParticipants [] = []
pairParticipants [_] = []
pairParticipants (a : b : rest) = (a, b) : pairParticipants rest

-- case of
evenNumber :: [String] -> Bool
evenNumber participants = case participants of
  [] -> True
  [_] -> False
  (_ : _ : rest) -> evenNumber rest

startsWithB :: String -> Bool
startsWithB word = case word of
  ('B' : _) -> True
  _ -> False

-- let
presentValue :: String -> Int
presentValue present =
  let valuePerCharacter 'B' = 10
      valuePerCharacter 'b' = 10
      valuePerCharacter _ = 1

      values = map valuePerCharacter present
   in sum values

-- where

birthdaySong :: String -> String
birthdaySong name = go 1
  where
    go 5 = ""
    go 3 = "Happy birthday dear " ++ name ++ "\n" ++ go 4
    go n = "Happy birthday to you" ++ go (n + 1)

-- map, filter
bPresents :: [String] -> [String]
bPresents presents = filter startsWithB presents

getValues :: [String] -> [Int]
getValues presents = map presentValue presents

-- Lambda expressions
goodStuff :: [String] -> [String]
goodStuff presents = filter (\p -> presentValue p >= 15) presents

friendshipLevel :: [String] -> [String] -> [(String, Int)]
friendshipLevel invited presents =
  zipWith (\person present -> (person, presentValue present)) invited presents

blackList :: [String] -> [String] -> [(String, String)]
blackList invited presents =
  let pairs = zip invited presents
   in filter (\(_, present) -> presentValue present < 15) pairs

-- We won't introduce the following concepts now, but look them up
--
-- Guards
-- Combining guards and pattern matching
-- List comprehensions
-- Data types
-- Type synonyms
-- Type annotations
