import Control.Monad (liftM, liftM2, (<=<))

{-
    Mini Course: Functional Programming
    Jeremie Gillet

    Day 2: Recursion
-}

factorial :: Integer -> Integer
factorial 0 = 1
factorial 1 = 1
factorial n = n * factorial (n - 1)

fibonacci :: Integer -> Integer
fibonacci 1 = 1
fibonacci 2 = 1
fibonacci n = fibonacci (n - 1) + fibonacci (n - 2)

chunks :: Int -> [a] -> [[a]]
chunks _ [] = []
chunks n list = take n list : chunks n (drop n list)

{-
    Day 2: Tail recursion
-}

factorial' :: Integer -> Integer
factorial' n = fac n 1
  where
    fac 0 product = product
    fac n product = fac (n - 1) (n * product)

{-
    Day 2: Folds
-}

sum' :: (Num a) => [a] -> a
sum' = foldr (+) 0

{-
    Day 2: Maybe
-}

find :: (Eq a) => (a -> Bool) -> [a] -> Maybe a
find _ [] = Nothing
find f (x : xs) = if f x then Just x else find f xs

safeDiv :: Int -> Int -> Maybe Int
safeDiv _ 0 = Nothing
safeDiv x y = Just (div x y)

safeHead :: [a] -> Maybe a
safeHead [] = Nothing
safeHead (x : _) = Just x

findEvenNumberAndDivideBy2 :: [Int] -> Maybe Int
findEvenNumberAndDivideBy2 x = case find even x of
  Nothing -> Nothing
  Just k -> Just (div k 2)

findEvenNumberAndDivideBy2' :: [Int] -> Maybe Int
findEvenNumberAndDivideBy2' = liftM (flip div 2) . find even

findEvenNumberAndDivideBynThen2n :: Int -> [Int] -> Maybe Int
findEvenNumberAndDivideBynThen2n n x = case find even x of
  Nothing -> Nothing
  Just k -> case safeDiv k n of
    Nothing -> Nothing
    Just k' -> safeDiv k' (2 * n)

findEvenNumberAndDivideBynThen2n' :: Int -> [Int] -> Maybe Int
findEvenNumberAndDivideBynThen2n' n = flip safeDiv (2 * n) <=< flip safeDiv n <=< find even

multiplyBothHeads :: [Int] -> [Int] -> Maybe Int
multiplyBothHeads as bs =
  let a = safeHead as
      b = safeHead bs
   in case (a, b) of
        (Just x, Just y) -> Just (x * y)
        _ -> Nothing

multiplyBothHeads' :: [Int] -> [Int] -> Maybe Int
multiplyBothHeads' as bs = liftM2 (*) (safeHead as) (safeHead bs)
