{-

    Example: Either instead of Maybe

-}

safeDiv :: Int -> Int -> Maybe Int
safeDiv _ 0 = Nothing
safeDiv x y = Just (div x y)

safeHead :: [a] -> Maybe a
safeHead [] = Nothing
safeHead (x : _) = Just x


safeDiv' :: Int -> Int -> Either String Int
safeDiv' _ 0 = Left "Error: division by 0"
safeDiv' x y = Right (div x y)

safeHead' :: [a] -> Either String a
safeHead' [] = Left "Error: head of an empty list"
safeHead' (x : _) = Right x

{-

    Maybe-specific Functions

-}

data Maybe' a = Nothing' | Just' a

(~>>=) :: Maybe' a -> (a -> Maybe' b) -> Maybe' b
Nothing' ~>>= f = Nothing'
(Just' x) ~>>= f = f x


return' :: a -> Maybe' a
return' = Just'

{-

    General Monadic Functions

-}

join :: Monad m => m (m a) -> m a
join mma = mma >>= id


(<=<) :: Monad m => (b -> m c) -> (a -> m b) -> a -> m c
(g <=< f) a = f a >>= g


liftM :: Monad m => (a -> b) -> m a -> m b
liftM f ma = ma >>= (return . f)


liftM2 :: Monad m => (a -> b -> c) -> m a -> m b -> m c
liftM2 f ma mb = ma >>= (\a ->
                 mb >>= (\b ->
                 return (f a b)))

liftM2' f ma mb = do
                  a <- ma
                  b <- mb
                  return (f a b)

