import Control.Monad.Reader

{-

    Custom Reader Monad

-}

data Reader' e a = Reader' (e -> a)

reader' :: (e -> a) -> Reader' e a
reader' = Reader'

runReader' :: Reader' e a -> e -> a
runReader' (Reader' r) = r

(~>>=) :: Reader' e a -> (a -> Reader' e b) -> Reader' e b
--(>>=) :: (e -> a) -> (a -> e -> b) -> e -> b
(Reader' r) ~>>= f = Reader' (\e -> let Reader' r1 = (f.r) e in r1 e)
-- r >>= f = \e -> (f.r) e e

return' :: a -> Reader' e a
return' x = Reader' (const x)


{-

    Example

-}

zoo = ["Lion", "Giraffe", "Hippopotamus", "Tiger", "Monkey"]

numOfAnimals :: Reader [String] Int
numOfAnimals = reader length

isPredator :: String -> Bool
isPredator "Lion" = True
isPredator "Tiger" = True
isPredator _ = False

numOfPredators :: Reader [String] Int
numOfPredators = reader (\e -> length $ filter isPredator e)

predatorsRatio :: Reader [String] Float
predatorsRatio = liftM2 (/) fnP fnA
                 where
                    fnP = liftM fromIntegral numOfPredators
                    fnA = liftM fromIntegral numOfAnimals
