import Control.Monad.State

{-

    Custom State Monad

-}

data State' s a = State' (s -> (a, s))

state' :: (s -> (a, s)) -> State' s a
state' = State'

runState' :: State' s a -> s -> (a, s)
runState' (State' act) = act

evalState' :: State' s a -> s -> a
evalState' (State' act) = fst . act

(~>>=) :: State' s a -> (a -> State' s b) -> State' s b
-- (~>>=) :: (s -> (s, a)) -> (a -> s -> (s, b)) -> s -> (s,b)
(State' act) ~>>= f = State' (\s -> let (a, s1) = act s
                                        State' act1 = f a
                                     in act1 s1)

return' :: a -> State' s a
return' x = State' (\s -> (x, s))


{-

    Example: Pseudo-random numbers

-}

type Random a = State Int a

bsd_rand :: Int -> Int
bsd_rand n = (n * 1103515245 + 12345) `mod` 2^31

sample :: (Int -> a) -> Random a
sample f = state (\s -> let rand_int = bsd_rand s in
                            (f rand_int, rand_int))


rand_choice :: [a] -> Random a
rand_choice xs = sample (\i -> xs !! (i `mod` length xs))


rand01 :: Random Float
rand01 = sample (\s -> fromIntegral s / (2^31))


rand_add :: Num a => Random a -> Random a -> Random a
rand_add = liftM2 (+)


uniform_mixture :: [(Float, Float)] -> Random Float
uniform_mixture gs = do 
                     (a, b) <- rand_choice gs
                     x <- rand01
                     return (a + x*(b - a))

