import Control.Monad.Writer

{-

Custom Writer Monad

-}

data Writer' w a = Writer' a w


writer' :: (a, w) -> Writer' w a
writer' (a, w) = Writer' a w


execWriter' :: Writer' w a -> w
execWriter' (Writer' _ w) = w


(~>>=) :: Writer' String a -> (a -> Writer' String b) -> Writer' String b
--(~>>=) :: (a, String) -> (a -> (b, String)) -> (b, String)
(Writer' x s) ~>>= f = let (Writer' y s') = f x in
                     Writer' y (s ++ s')

return' x = Writer' x ""

{-

    Example: Logging computation of the collatz sequence

-}

collatz :: Int -> Int
collatz n = if n `mod` 2 == 0
            then z1
            else z2
                where
                    z1 = n `div` 2
                    z2 = 3*n + 1


collatzSeq :: Int -> [Int]
collatzSeq = iterate collatz


iterate' :: (a -> a) -> a -> [a]
iterate' f a = a : iterate' f (f a)


iterateM :: Monad m => (a -> m a) -> a -> m [a]
iterateM f a = liftM (a:) ((f a) >>= iterateM f)


collatzWriter :: Int -> Writer String Int
collatzWriter n = if n `mod` 2 == 0
                  then (writer (z1, sz1))
                  else (writer (z2, sz2))
                    where
                        z1 = n `div` 2
                        sz1 = show n ++ " is even => " ++ show n ++ "/2 = " ++ show z1 ++ "\n"

                        z2 = 3*n + 1
                        sz2 = show n ++ " is odd => " ++ "3*" ++ show n ++ " + 1 = " ++ show z2 ++ "\n"

collatzSeqWriter :: Int -> Writer String [Int]
collatzSeqWriter = iterateM collatzWriter



